@@ -355,7 +355,6 @@ int array_new(b2nd_context_t *ctx, int special_value, b2nd_array_t **array) {
   if ((*array)->nitems != 0) {
     int64_t nchunks = (*array)->extnitems / (*array)->chunknitems;
     int64_t nitems = nchunks * (*array)->extchunknitems;
-    // blosc2_schunk_fill_special(sc, nitems, BLOSC2_SPECIAL_ZERO, chunksize);
     BLOSC_ERROR(blosc2_schunk_fill_special(sc, nitems, special_value, chunksize));
   }
   (*array)->sc = sc;
@@ -395,6 +394,23 @@ int b2nd_zeros(b2nd_context_t *ctx, b2nd_array_t **array) {
 }
 
 
+int b2nd_nans(b2nd_context_t *ctx, b2nd_array_t **array) {
+  BLOSC_ERROR_NULL(ctx, BLOSC2_ERROR_NULL_POINTER);
+  BLOSC_ERROR_NULL(array, BLOSC2_ERROR_NULL_POINTER);
+
+  BLOSC_ERROR(array_new(ctx, BLOSC2_SPECIAL_NAN, array));
+
+  const int32_t typesize = (*array)->sc->typesize;
+  if (typesize != 4 && typesize != 8)
+  {
+    BLOSC_TRACE_ERROR("Unsupported typesize for NaN");
+    return BLOSC2_ERROR_DATA;
+  }
+
+  return BLOSC2_ERROR_SUCCESS;
+}
+
+
 int b2nd_full(b2nd_context_t *ctx, b2nd_array_t **array, const void *fill_value) {
   BLOSC_ERROR_NULL(ctx, BLOSC2_ERROR_NULL_POINTER);
   BLOSC_ERROR_NULL(array, BLOSC2_ERROR_NULL_POINTER);
@@ -689,6 +705,10 @@ int64_t nchunk_fastpath(const b2nd_array_t *array, const int64_t *start,
   int ndim = (int) array->ndim;
   int inner_dim = ndim - 1;
   int64_t partial_slice_size = 1;
+  int outer_dims = 0;
+  for (int j = 0; j < inner_dim; ++j) {
+    outer_dims += (array->blockshape[j] != 1);
+   }
   int64_t partial_chunk_size = 1;
   for (int i = ndim - 1; i >= 0; --i) {
     // We need to check if the slice is contiguous in the C order
@@ -707,11 +727,13 @@ int64_t nchunk_fastpath(const b2nd_array_t *array, const int64_t *start,
         }
       }
       else {
-        if (array->chunkshape[i] != array->blockshape[i]) {
+        // A block is still contiguous in the C order if the outer dimensions are 1 and the inner is
+        // a divisor of the chunkshape
+        if ( ! (array->chunkshape[i] == array->blockshape[i] ||
+                (outer_dims == 0 && array->chunkshape[i] % array->blockshape[i] == 0))) {
           return -1;
         }
       }
-      inner_dim = i;
     }
 
     // We need start and stop to be aligned with the chunkshape
@@ -726,7 +748,6 @@ int64_t nchunk_fastpath(const b2nd_array_t *array, const int64_t *start,
       return -1;
     }
   }
-
   // Compute the chunk number
   int64_t *chunks_idx;
   int nchunks = b2nd_get_slice_nchunks(array, start, stop, &chunks_idx);
@@ -754,11 +775,7 @@ int get_set_slice(void *buffer, int64_t buffersize, const int64_t *start, const
     BLOSC_ERROR(BLOSC2_ERROR_INVALID_PARAM);
   }
 
-  uint8_t *buffer_b = (uint8_t *) buffer;
-  const int64_t *buffer_start = start;
-  const int64_t *buffer_stop = stop;
-  const int64_t *buffer_shape = shape;
-
+  uint8_t *buffer_b = buffer;
   int8_t ndim = array->ndim;
 
   // 0-dim case
@@ -863,11 +880,11 @@ int get_set_slice(void *buffer, int64_t buffersize, const int64_t *start, const
   int64_t update_nchunks = 1;
   for (int i = 0; i < ndim; ++i) {
     int64_t pos = 0;
-    while (pos <= buffer_start[i]) {
+    while (pos <= start[i]) {
       pos += array->chunkshape[i];
     }
     update_start[i] = pos / array->chunkshape[i] - 1;
-    while (pos < buffer_stop[i]) {
+    while (pos < stop[i]) {
       pos += array->chunkshape[i];
     }
     update_shape[i] = pos / array->chunkshape[i] - update_start[i];
@@ -895,7 +912,7 @@ int get_set_slice(void *buffer, int64_t buffersize, const int64_t *start, const
     }
     bool chunk_empty = false;
     for (int i = 0; i < ndim; ++i) {
-      chunk_empty |= (chunk_stop[i] <= buffer_start[i] || chunk_start[i] >= buffer_stop[i]);
+      chunk_empty |= (chunk_stop[i] <= start[i] || chunk_start[i] >= stop[i]);
     }
     if (chunk_empty) {
       continue;
@@ -906,7 +923,7 @@ int get_set_slice(void *buffer, int64_t buffersize, const int64_t *start, const
       // Check if all the chunk is going to be updated and avoid the decompression
       bool decompress_chunk = false;
       for (int i = 0; i < ndim; ++i) {
-        decompress_chunk |= (chunk_start[i] < buffer_start[i] || chunk_stop[i] > buffer_stop[i]);
+        decompress_chunk |= (chunk_start[i] < start[i] || chunk_stop[i] > stop[i]);
       }
 
       if (decompress_chunk) {
@@ -1001,8 +1018,8 @@ int get_set_slice(void *buffer, int64_t buffersize, const int64_t *start, const
       // compute the start of the slice inside the block
       int64_t slice_start[B2ND_MAX_DIM] = {0};
       for (int i = 0; i < ndim; ++i) {
-        if (block_start[i] < buffer_start[i]) {
-          slice_start[i] = buffer_start[i] - block_start[i];
+        if (block_start[i] < start[i]) {
+          slice_start[i] = start[i] - block_start[i];
         } else {
           slice_start[i] = 0;
         }
@@ -1011,8 +1028,8 @@ int get_set_slice(void *buffer, int64_t buffersize, const int64_t *start, const
 
       int64_t slice_stop[B2ND_MAX_DIM] = {0};
       for (int i = 0; i < ndim; ++i) {
-        if (block_stop[i] > buffer_stop[i]) {
-          slice_stop[i] = block_shape[i] - (block_stop[i] - buffer_stop[i]);
+        if (block_stop[i] > stop[i]) {
+          slice_stop[i] = block_shape[i] - (block_stop[i] - stop[i]);
         } else {
           slice_stop[i] = block_stop[i] - block_start[i];
         }
@@ -1025,13 +1042,12 @@ int get_set_slice(void *buffer, int64_t buffersize, const int64_t *start, const
       }
 
       uint8_t *src = &buffer_b[0];
-      const int64_t *src_pad_shape = buffer_shape;
 
       int64_t src_start[B2ND_MAX_DIM] = {0};
       int64_t src_stop[B2ND_MAX_DIM] = {0};
       for (int i = 0; i < ndim; ++i) {
-        src_start[i] = slice_start[i] - buffer_start[i];
-        src_stop[i] = slice_stop[i] - buffer_start[i];
+        src_start[i] = slice_start[i] - start[i];
+        src_stop[i] = slice_stop[i] - start[i];
       }
 
       uint8_t *dst = &data[nblock * array->blocknitems * array->sc->typesize];
@@ -1048,13 +1064,13 @@ int get_set_slice(void *buffer, int64_t buffersize, const int64_t *start, const
       }
 
       if (set_slice) {
-        b2nd_copy_buffer(ndim, array->sc->typesize,
-                         src, src_pad_shape, src_start, src_stop,
-                         dst, dst_pad_shape, dst_start);
+        b2nd_copy_buffer2(ndim, array->sc->typesize,
+                          src, shape, src_start, src_stop,
+                          dst, dst_pad_shape, dst_start);
       } else {
-        b2nd_copy_buffer(ndim, array->sc->typesize,
-                         dst, dst_pad_shape, dst_start, dst_stop,
-                         src, src_pad_shape, src_start);
+        b2nd_copy_buffer2(ndim, array->sc->typesize,
+                          dst, dst_pad_shape, dst_start, dst_stop,
+                          src, shape, src_start);
       }
     }
 
@@ -1309,6 +1325,161 @@ int b2nd_copy(b2nd_context_t *ctx, const b2nd_array_t *src, b2nd_array_t **array
 }
 
 
+int b2nd_concatenate(b2nd_context_t *ctx, const b2nd_array_t *src1, const b2nd_array_t *src2,
+                     int8_t axis, bool copy, b2nd_array_t **array) {
+  BLOSC_ERROR_NULL(src1, BLOSC2_ERROR_NULL_POINTER);
+  BLOSC_ERROR_NULL(src2, BLOSC2_ERROR_NULL_POINTER);
+  BLOSC_ERROR_NULL(array, BLOSC2_ERROR_NULL_POINTER);
+
+  // Validate the axis parameter
+  if (axis < 0 || axis >= src1->ndim) {
+    BLOSC_TRACE_ERROR("axis parameter is out of bounds: axis=%d, expected range=[0, %d)", axis, src1->ndim - 1);
+    BLOSC_ERROR(BLOSC2_ERROR_INVALID_PARAM);
+  }
+
+  // typesize must be the same for both arrays
+  if (src1->sc->typesize != src2->sc->typesize) {
+    BLOSC_TRACE_ERROR("The two arrays must have the same typesize");
+    BLOSC_ERROR(BLOSC2_ERROR_INVALID_PARAM);
+  }
+
+  // Keep the src1 shape for later use
+  int64_t src1_shape[B2ND_MAX_DIM];
+  for (int i = 0; i < src1->ndim; ++i) {
+    src1_shape[i] = src1->shape[i];
+  }
+
+  // Support for 0-dim arrays is not implemented
+  if (src1->ndim == 0 || src2->ndim == 0) {
+    BLOSC_TRACE_ERROR("Concatenation of 0-dim arrays is not supported");
+    BLOSC_ERROR(BLOSC2_ERROR_INVALID_PARAM);
+  }
+
+  // Check that the shapes are compatible for concatenation
+  if (src1->ndim != src2->ndim) {
+    BLOSC_TRACE_ERROR("The two arrays must have the same number of dimensions");
+    BLOSC_ERROR(BLOSC2_ERROR_INVALID_PARAM);
+  }
+  // Compute the new shape
+  int64_t newshape[B2ND_MAX_DIM];
+  for (int8_t i = 0; i < src1->ndim; ++i) {
+    if (i == axis) {
+      newshape[i] = src1->shape[i] + src2->shape[i];
+    } else {
+      if (src1->shape[i] != src2->shape[i]) {
+        BLOSC_TRACE_ERROR("The two arrays must have the same shape in all dimensions except the concatenation axis");
+        BLOSC_ERROR(BLOSC2_ERROR_INVALID_PARAM);
+      }
+      newshape[i] = src1->shape[i];
+    }
+  }
+
+  if (copy) {
+    BLOSC_ERROR(b2nd_copy(ctx, src1, array));
+  }
+  else {
+    *array = (b2nd_array_t *)src1;
+  }
+
+  // Extend the array, we don't need to specify the start in resize, as we are extending the shape from the end
+  BLOSC_ERROR(b2nd_resize(*array, newshape, NULL));
+
+  // Copy the data from the second array
+  int64_t start[B2ND_MAX_DIM];
+  int64_t stop[B2ND_MAX_DIM];
+
+  // Check if the chunk is aligned with dest chunks, and has the same blockshape
+  bool aligned = true;
+  for (int8_t i = 0; i < src2->ndim; ++i) {
+    if (src1->chunkshape[i] != src2->chunkshape[i] ||
+        src2->blockshape[i] != (*array)->blockshape[i] ||
+        (i == axis && (src1_shape[i]) % (*array)->chunkshape[i] != 0)
+        ) {
+      aligned = false;
+      break;
+        }
+  }
+  // ...and get the chunk index in the dest array if aligned
+  int64_t chunks_in_array_strides[B2ND_MAX_DIM];
+  // Calculate strides for destination array
+  chunks_in_array_strides[(*array)->ndim - 1] = 1;
+  for (int i = (*array)->ndim - 2; i >= 0; --i) {
+    chunks_in_array_strides[i] = chunks_in_array_strides[i + 1] *
+                                ((*array)->extshape[i + 1] / (*array)->chunkshape[i + 1]);
+  }
+
+  // Copy chunk by chunk
+  void *buffer = malloc(src2->sc->typesize * src2->extchunknitems);
+  BLOSC_ERROR_NULL(buffer, BLOSC2_ERROR_MEMORY_ALLOC);
+  for (int64_t nchunk = 0; nchunk < src2->sc->nchunks; ++nchunk) {
+    // Get multidimensional chunk position
+    int64_t nchunk_ndim[B2ND_MAX_DIM] = {0};
+    int64_t chunkshape[B2ND_MAX_DIM] = {0};
+    for (int8_t i = 0; i < src2->ndim; ++i) {
+      chunkshape[i] = src2->chunkshape[i];
+    }
+    int64_t chunks_in_dim[B2ND_MAX_DIM] = {0};
+    for (int8_t i = 0; i < src2->ndim; ++i) {
+      chunks_in_dim[i] = src2->extshape[i] / src2->chunkshape[i];
+    }
+    blosc2_unidim_to_multidim(src2->ndim, chunks_in_dim, nchunk, nchunk_ndim);
+
+    if (aligned) {
+      // Get the uncompressed chunk buffer from the source array
+      bool needs_free = false;
+      uint8_t *chunk;
+      int32_t cbytes = blosc2_schunk_get_chunk(src2->sc, nchunk, &chunk, &needs_free);
+      if (cbytes < 0) {
+        BLOSC_TRACE_ERROR("Error getting chunk from source array");
+        BLOSC_ERROR(BLOSC2_ERROR_FAILURE);
+      }
+      // Update the chunk in the destination array
+      // We need to free only if needs_free is true or copy is false
+      // bool needs_copy = !needs_free || copy;
+      // BLOSC_ERROR(blosc2_schunk_update_chunk((*array)->sc, nchunk_dest, chunk, needs_copy));
+      // if (needs_free && !copy) {
+      //   free(chunk);
+      // }
+      // TODO: the above makes some tests to crash, so always force a copy; try to optimize this later
+      int64_t nchunk_dest = 0;
+      nchunk_ndim[axis] += src1_shape[axis] / (*array)->chunkshape[axis];
+      for ( int i =0; i< src2->ndim; i++) {
+        nchunk_dest += nchunk_ndim[i] * chunks_in_array_strides[i];
+      }
+      BLOSC_ERROR(blosc2_schunk_update_chunk((*array)->sc, nchunk_dest, chunk, true));
+      if (needs_free) {
+        free(chunk);
+      }
+    }
+    else {
+
+      // Set positions for each dimension
+      for (int8_t i = 0; i < src2->ndim; ++i) {
+        start[i] = nchunk_ndim[i] * src2->chunkshape[i];
+        stop[i] = start[i] + src2->chunkshape[i];
+        if (stop[i] > src2->shape[i]) {
+          stop[i] = src2->shape[i];  // Handle boundary chunks
+        }
+      }
+      // Load chunk into buffer
+      BLOSC_ERROR(b2nd_get_slice_cbuffer(src2, start, stop, buffer, chunkshape, src2->sc->chunksize));
+
+      // Apply chunk offset only for concatenation axis
+      start[axis] += src1_shape[axis];
+      stop[axis] += src1_shape[axis];
+
+      // Copy the chunk to the correct position
+      BLOSC_ERROR(b2nd_set_slice_cbuffer(buffer, chunkshape,
+                                         src2->sc->typesize * src2->extchunknitems,
+                                         start, stop, *array));
+    }
+  }
+
+  free(buffer);
+
+  return BLOSC2_ERROR_SUCCESS;
+}
+
 int b2nd_save(const b2nd_array_t *array, char *urlpath) {
   BLOSC_ERROR_NULL(array, BLOSC2_ERROR_NULL_POINTER);
   BLOSC_ERROR_NULL(urlpath, BLOSC2_ERROR_NULL_POINTER);
@@ -2082,13 +2253,15 @@ b2nd_create_ctx(const blosc2_storage *b2_storage, int8_t ndim, const int64_t *sh
   }
 
   if (dtype == NULL) {
-    ctx->dtype = strdup(B2ND_DEFAULT_DTYPE);
-    ctx->dtype_format = 0;  // The default is NumPy format
+    // ctx->dtype = strdup(B2ND_DEFAULT_DTYPE);
+    char buf[16] = {0};
+    snprintf(buf, sizeof(buf), "|S%d", cparams->typesize);
+    ctx->dtype = strdup(buf);
   }
   else {
     ctx->dtype = strdup(dtype);
-    ctx->dtype_format = dtype_format;
   }
+  ctx->dtype_format = dtype_format;
 
   params_b2_storage->cparams = cparams;
   ctx->b2_storage = params_b2_storage;
